/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraqua
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"

/****************************************************************************/
/* external variables */

extern descript ws_des;
extern robot therobot;
extern FILE *log_fd;
extern long nb_hit_calls;
extern long nb_goodson_calls;

/****************************************************************************
  Function to check if the line segment is collision free 
*****************************************************************************/

boolean hit_seg(fpoint3D *X,image V,int a,int b){
  
  register int i;
  int count,jump;
  double hit_max;
  fpoint3D hit_init,hit_cur,hit_delta;
  
  hit_init.x = X[a].x*ws_des.dimx;
  hit_init.y = X[a].y*ws_des.dimy;
  hit_init.z = X[a].z*ws_des.dimz;
  hit_delta.x = ws_des.dimx*(X[b].x - X[a].x);
  hit_delta.y = ws_des.dimy*(X[b].y - X[a].y);
  hit_delta.z = ws_des.dimz*(X[b].z - X[a].z);
  hit_max = MAX(fabs(hit_delta.y),fabs(hit_delta.z)); 
  hit_max = MAX(hit_max,fabs(hit_delta.x));
  count=(int)(hit_max)+1;
  hit_delta.x /= (float)count;
  hit_delta.y /= (float)count;
  hit_delta.z /= (float)count;
  hit_cur=hit_init;

  for(i=1;i<=count;){
    if((jump=V.d[(int)hit_cur.x][(int)hit_cur.y][(int)hit_cur.z])
       ==OBSTACLE){
      return(TRUE);
    }
    if (jump < 3)
      i++;
    else
      i+=(jump-1);
    hit_cur.x = hit_init.x+(hit_delta.x*i); 
    hit_cur.y = hit_init.y+(hit_delta.y*i); 
    hit_cur.z = hit_init.z+(hit_delta.z*i);
  }
  return(FALSE);
}

/****************************************************************************
  Function to check if all the links are collision free 
*****************************************************************************/

boolean hit(fpoint3D *X,image V) {

  register int i;
  
  nb_hit_calls++;
  for(i=0;i<therobot.nblink;i++) {
    if (hit_seg(X,V,therobot.robotlink[i].endp1,
		therobot.robotlink[i].endp2))
      return(TRUE);
  }
    
  return(FALSE);  
}

/****************************************************************************
  Function to compute Cross_product of two 3D vector
  To speedup, may be replaced by macros.
*****************************************************************************/

void cross_prod(fvector3D p1,fvector3D p2,fvector3D *p3) {

  p3->x=(p1.y)*(p2.z)-(p1.z)*(p2.y);
  p3->y=(p1.z)*(p2.x)-(p1.x)*(p2.z);
  p3->z=(p1.x)*(p2.y)-(p1.y)*(p2.x);
  return;

}
    
boolean seg_collide(fpoint3D p1,fpoint3D p2,fpoint3D p3,fpoint3D p4){

  fpoint3D v1,v2,u,w,w1,w2;
  double d1,d2,d11,d12,d21,d22,wmag,dist;

  subst_vec(p2,p1,v1);
  subst_vec(p4,p3,v2);
  subst_vec(p4,p1,u);

  cross_prod(v1,v2,&w);
  wmag=sqrt(w.x*w.x + w.y*w.y + w.z*w.z);
  if (wmag < 0.0000001) {
    cross_prod(v1,u,&w1);
    cross_prod(v1,w1,&w);
    wmag=sqrt(w.x*w.x + w.y*w.y + w.z*w.z);
    dist=dot_prod(u,w)/wmag;
    if (fabs(dist) > 2.0*GROW )
      return(FALSE);
    else
      return(TRUE);
  } else {
    cross_prod(v1,w,&w1);
    cross_prod(v2,w,&w2);

    d1=dot_prod(w1,p1);
    d2=dot_prod(w2,p3);
    d11=dot_prod(w1,p3)-d1;
    d12=dot_prod(w1,p4)-d1;
    d21=dot_prod(w2,p1)-d2;
    d22=dot_prod(w2,p2)-d2;

    if (((d11*d12) < 0.0)&&((d21*d22) < 0.0))  {
      dist=dot_prod(u,w)/wmag;
      if (fabs(dist) > (2.0*GROW))
        return(FALSE);
      else
        return(TRUE);
    }
    else
      return(FALSE);
  }
}

/****************************************************************************
  Function to see if the given configuration satisfies the given kinematic 
  constraint: 
  1. joint limit
  2. self-intersect
  This is to be more general, can save some time by replacing it by some 
  more example-specific routine.
****************************************************************************/
boolean goodson(configuration X,fpoint3D *cur_point) {
  
  register int i,j,dof=0;
  
  nb_goodson_calls++;
  /* check joint limits or mechanical stops */
  for(i=0;i<therobot.nbframe;i++) {
    for(j=0;j<6;j++) {
      if(therobot.localframe[i].dofmask & (1<<j)) {
	if (X[dof]>therobot.localframe[i].maxlimit[j]) 
	  return(FALSE);
	if (X[dof]<therobot.localframe[i].minlimit[j]) 
	  return(FALSE);
	dof++;
      }
    }
  }

  /* Make sure that the dof is cerrect */
  if (dof!=therobot.dof) {
    perror("goodson error\n");
    exit(-1);
  }
  
  /* Check if the links are intersecting each other */
  for(i=0;i<therobot.nblink;i++) {
    for(j=i+1;j<therobot.nblink;j++) {
      if (i==j) continue;
      if ((therobot.robotlink[i].endp1==therobot.robotlink[j].endp2) ||
	  (therobot.robotlink[i].endp2==therobot.robotlink[j].endp1))
	continue;
      
      if (seg_collide(cur_point[therobot.robotlink[i].endp1],
		      cur_point[therobot.robotlink[i].endp2],
		      cur_point[therobot.robotlink[j].endp1],
		      cur_point[therobot.robotlink[j].endp2]))
	return(FALSE);
    }
  }

  return(TRUE);
}


