/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>
#include <math.h>
#include <time.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"

/* local constant definition */

#define NBPROC       100
#define NBBROWN      200

/* external variables */

extern robot therobot;
extern descript ws_des;
extern FILE *log_fd;
extern unsigned short xsubi[3];
extern void bcopy();

/*****************************************************************************
  Function to find path (Main function call)
******************************************************************************/

int find_path(image V, image voro, path *finalpath) {

  boolean out,found=FALSE,jumped;
  int r,i,nbtrials,nbsteps;
  int valid_begin,valid_end,nbbrown=NBBROWN;
  int dimx2=ws_des.dimx*0.5,dimx=ws_des.dimx;
  path rand_path,good_path,cur_path;
  configuration curq,trialq;
  float Ucur, Unew;
  image heurpot[MAXLEAD];
  ipoint3D k_init_lead[MAXLEAD],k_goal_lead[MAXLEAD];
  fpoint3D k_init_point[MAXNBP];
  

#ifdef DEBUG
  fprintf(log_fd,"initial point conf: ");
  for (i=0;i<therobot.dof;i++) 
    fprintf(log_fd,"%3.2f,",therobot.initq[i]);
  NL(log_fd);
#endif  

#ifdef VERBOSE
  if (verbose) {
    fprintf(stderr,"initial point conf: ");
    for (i=0;i<therobot.dof;i++) 
      fprintf(stderr,"%3.2f,",therobot.initq[i]);
    NL(stderr);
  }
#endif  

  FK(therobot.initq,k_init_point);

#ifdef DEBUG
  for (i=0;i<therobot.nbpoint;i++) 
    fprintf(log_fd,"beginning endpoint %d: %3.2f, %3.2f, %3.2f\n",
	    i,k_init_point[i].x,k_init_point[i].y,k_init_point[i].z);
#endif
#ifdef VERBOSE
  if (verbose) {
    for (i=0;i<therobot.nbpoint;i++) 
      fprintf(stderr,"beginning endpoint %d: %3.2f, %3.2f, %3.2f\n",
	      i,k_init_point[i].x,k_init_point[i].y,k_init_point[i].z);
  }
#endif

  outside(k_init_point,out);
  
  if (out) {
    perror("begin configuration outside frame!");
    exit(-1);
  }
  
  valid_begin=HIT(k_init_point,V);

  if(valid_begin) {
    perror("begin configuration inside obstacles!");
    exit(-1);
  }
  
  trans_lead(k_init_lead);

  valid_end = 0;

  for(i=0;i<therobot.nblead;++i){
    k_goal_lead[i].x = (int)(therobot.goallead[i].x * ws_des.dimx);
    k_goal_lead[i].y = (int)(therobot.goallead[i].y * ws_des.dimy);
    k_goal_lead[i].z = (int)(therobot.goallead[i].z * ws_des.dimz);
#ifdef DEBUG
    fprintf(log_fd,"control point %d at goal: %d, %d, %d\n",
	    i,k_goal_lead[i].x,k_goal_lead[i].y,k_goal_lead[i].z);
#endif
#ifdef VERBOSE
    if (verbose) {
      fprintf(stderr,"control point %d at goal: %d, %d, %d\n",
	      i,k_goal_lead[i].x,k_goal_lead[i].y,k_goal_lead[i].z);
    }
#endif
    valid_end = valid_end || (V.d[k_goal_lead[i].x][k_goal_lead[i].y]
			    [k_goal_lead[i].z] == OBSTACLE);
  }
  
  if(valid_end) {
    perror("goal position hitting obstacles");
    exit(-1);
  }
  
  
  /* Computation of the heuristic potential */
  
  fprintf(stderr,"Computing the heuristic potential\n");

  /* Main function to compute heuristic functions */
  heuristic(voro.d,V.d,&(k_goal_lead[0]),&(heurpot[0]));
  
  if(heurpot[0].d[k_init_lead[0].x][k_init_lead[0].y][k_init_lead[0].z] 
     == UNVISITED){
    perror("freespace not connected"); 
    exit(-1);
  }

#ifdef DEBUG
  for (i=0;i<therobot.nblead;i++) {
    fprintf(log_fd,"goal %d = %d\n ",i+1,
	    heurpot[i].d[k_goal_lead[i].x][k_goal_lead[i].y][k_goal_lead[i].z]);
  }
#endif

  /* Computing the robot path */

  fprintf(stderr,"Computing the robot path ...\n");
  
#ifdef VERBOSE
  if (verbose) 
    fprintf(stderr,"Allocating momery ...\n");
#endif

  /* allocate memory for the paths */
  rand_path.d = (configuration *)calloc(2*NBBROWN*NBBROWN,
					sizeof(configuration));
  good_path.d = (configuration *)calloc(2*NBBROWN,sizeof(configuration));
  cur_path.d = (configuration *)calloc(3*NBBROWN*NBBROWN,
				       sizeof(configuration));

  assign(therobot.initq,curq);
  nbtrials = 0;
  cur_path.dim = 1;
  
  found = down(curq,&Ucur,good_path.d,&good_path.dim,V,heurpot);

#ifdef DEBUG
  fprintf(log_fd,"After %d move, down to :",good_path.dim);
  for (i=0;i<therobot.dof;i++) 
    fprintf(log_fd,"%3.2f,",good_path.d[good_path.dim-1][i]);
  NL(log_fd);
#endif

#ifdef VERBOSE
  if (verbose) {
    fprintf(stderr,"After %d move, down to :",good_path.dim);
    for (i=0;i<therobot.dof;i++) 
      fprintf(stderr,"%3.2f,",good_path.d[good_path.dim-1][i]);
    NL(stderr);
  }
#endif

  concat(good_path,&cur_path);
  
  /* curq is modified by the program */
  if(!found) {
    fprintf(stderr,"reached a local minimum at altitude %f\n",Ucur);
#ifdef DEBUG
    fprintf(log_fd,"reached a local minimum at altitude %f\n",Ucur);
#endif    
  }

  while(!found && nbtrials<100) {

    jumped = 0;
    
    for(r=0;r<NBPROC;++r){

      fprintf(stderr,"nbproc = %d\n",r+1);
      nbbrown=dimx+(dimx2/NBPROC)*r;
      nbsteps = random()%(nbbrown*nbbrown);
      assign(curq,trialq);

      /* Need to reset the potential value */
      Unew=Ucur;

#ifdef DEBUG
      fprintf(log_fd,"%d proposed random walk\n",nbsteps);
#endif
#ifdef VERBOSE
      if (verbose)  {
	fprintf(stderr,"%d proposed random walk\n",nbsteps);
	fprintf(stderr,"random motion ... \n");
      }
#endif
      
      /* random motion */
      if (!brownian(trialq,&Unew,rand_path.d,&rand_path.dim,&nbsteps,
		    V,heurpot)) {
#ifdef DEBUG
	fprintf(log_fd,"%d actual random walk\n",nbsteps);
#endif
#ifdef VERBOSE
	if (verbose)  {
	  fprintf(stderr,"%d actual random walk\n",nbsteps);
	  fprintf(stderr,"down motion ... \n");
	}
#endif
      }

      found = down(trialq, &Unew, good_path.d,&good_path.dim,V,heurpot);
#ifdef DEBUG
      fprintf(log_fd,"After %d move, down to (",good_path.dim);
      for(i=0;i<therobot.dof;i++) 
	fprintf(log_fd,"%3.2f,",good_path.d[good_path.dim-1][i]);
      fprintf(log_fd,")\n");
#endif
#ifdef VERBOSE
      if (verbose) {
	fprintf(stderr,"After %d move, down to (",good_path.dim);
	for(i=0;i<therobot.dof;i++) 
	  fprintf(stderr,"%3.2f,",good_path.d[good_path.dim-1][i]);
	fprintf(stderr,")\n");
      }
#endif

      
      if(Unew < Ucur) {
	jumped = 1;
	break;
      }
      if(r >= 30) {
	cur_path.dim = random()%cur_path.dim; 
	assign(cur_path.d[cur_path.dim-1],curq);
	Ucur = MAXFLOAT;
      }
    }

    if(jumped){
      assign(trialq,curq);
      concat(rand_path,&cur_path);
      concat(good_path,&cur_path);
      Ucur = Unew;
      fprintf(stderr,"reached a local minimum at altitude %f\n",Ucur);
    }
    if(found) break;
    nbtrials++;
  }
  finalpath->dim = cur_path.dim;
  finalpath->d = cur_path.d;
  if (found) {
    fprintf(stderr, "An admissible path has been found with length %d\n",
	    cur_path.dim);
  }
  if(!found){ 
    fprintf(stderr," There is no admissible path\n");
    finalpath->dim = 0;
  }
  free(good_path.d);
  free(rand_path.d);
  for(i=0;i<therobot.nblead;i++) 
    free(heurpot[i].d);

  return(TRUE);
}


/*****************************************************************************
  Function to concat two path into one
******************************************************************************/

int concat(path path1,path *path2) {

  int i;
  int newdim;
  /* We add path1 to path2 */
  
  newdim = path1.dim + path2->dim - 1;
  
  for(i= 0;i<path1.dim;++i){
    assign(path1.d[i],path2->d[i+path2->dim-1]);
  }
  path2->dim = newdim;
  return(TRUE);

}

