/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <values.h>
#include <sys/file.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "solidp.h"
#include "parse.h"

/**************************************************************************
  declare local constants
***************************************************************************/

#define MIN_PATH_LENGTH 256

/**************************************************************************
  declare global variables
***************************************************************************/

unsigned short xsubi[3];
robot therobot;
descript ws_des;
FILE *log_fd;
int verbose;
long nb_fk_calls=0;
long nb_hit_calls=0;
long nb_goodson_calls=0;
long nb_no_hit=0;

/**************************************************************************
  Main function for solid (3D motion planning)
***************************************************************************/

int main(int argc,char **argv) {

  char bmap_name[80],seg_name[80],path_name[80],robot_name[80];
  FILE *fd_bmap,*fd_seg,*fd_path,*fd_robot;
  descript path_des;
  image bmap,V,voro;
  path finalpath;
  clock_t start,finish;
  int i,n,dimx,dimy,dimz;
  segment *world_seg;
  int world_nbseg;
  struct arg *args;
  int curvline=0;
#ifdef DEBUG
  fpoint3D qp[MAXNBP];
#endif
  int reseeding=0; /* nonzero if RNG state will be re-initialized */
  int seed=1; /* new seed for RNG */

  args=parse_args(argc,argv);

  if (args[chindex('h')].ch) {
    PrintHelp();
    exit(0);
  }

  if (args[chindex('v')].ch) 
    verbose=1;
  else
    verbose=0;
  

  if (args[chindex('c')].ch) {
    curvline=1;
  }

  if (!GetStrOpt(args,'i',bmap_name,"bmap.dat",0,12)) 
    fprintf(stderr,"Use default workspace bitmap file name %s\n",bmap_name);
  if (!GetStrOpt(args,'p',path_name,"path.dat",0,12)) 
    fprintf(stderr,"Use default path file name %s\n",path_name);
  if (!GetStrOpt(args,'r',robot_name,"robot.dat",0,12)) 
    fprintf(stderr,"Use default robot data file name %s\n",robot_name);
  if (!GetIntOpt(args,'n',&n,0,(-1),10)) 
    fprintf(stderr,"Use default action: without smoothing\n");
  if (!GetIntOpt(args,'s',&seed, getpid() ,0, (MAXINT)))
    reseeding=0;
  else 
    reseeding=1;

  fprintf(stderr,"Using seed:%d\n",seed);
    
  if(!(fd_bmap = fopen(bmap_name,"r"))){
    fprintf(stderr,"Cannot open file %s\n",bmap_name);
    exit(-1);
  }
    
  if(!(fd_robot = fopen(robot_name,"r"))){
    fprintf(stderr,"Cannot open file %s\n",robot_name);
    exit(-1);
  }
    
  if(!(fd_path = fopen(path_name,"w"))){
    fprintf(stderr,"Cannot open file %s\n",path_name);
    exit(-1);
  }

  strcpy(seg_name,bmap_name);
  strcat(seg_name,".seg");

  if(!(fd_seg = fopen(seg_name,"r")))
    fprintf(stderr,"Cannot open file %s\n",seg_name);	
    
  fread(&ws_des,sizeof(descript),1,fd_bmap);
  
  if(ws_des.octet != 1) {
    fprintf(stderr,"This is not a 1 byte coding!\n");
    exit(-1);
  }
    
  dimx = bmap.dimx = V.dimx = voro.dimx = ws_des.dimx;
  dimy = bmap.dimy = V.dimy = voro.dimy = ws_des.dimy;
  dimz = bmap.dimz = V.dimz = voro.dimz = ws_des.dimz;
  
  fread(&world_nbseg,sizeof(int),1,fd_seg);

  /* allocate memory for 3D bitmap:V */
  if((V.d = (unsigned char ***)calloc3D(dimx,dimy,dimz,
				     sizeof(unsigned char))) == NULL) {
    perror("Memory allocation error");
    exit(-1);
  }

  /* allocate memory for 3D bitmap:voro */
  if((voro.d = (unsigned char ***)calloc3D(dimx,dimy,dimz,
				     sizeof(unsigned char))) == NULL) {
    perror("Memory allocation error");
    exit(-1);
  }

  start=Start_Clock("read in data");

  /* allocate memory for world segments */
  world_seg= (segment *)calloc(world_nbseg,sizeof(segment));

  fread(world_seg,sizeof(segment),world_nbseg,fd_seg);
  fread(&(V.d[0][0][0]),sizeof(unsigned char),dimx*dimy*dimz,fd_bmap);
  fread(&(voro.d[0][0][0]),sizeof(unsigned char),dimx*dimy*dimz,fd_bmap);

  finish=Finish_Clock("read in data");
  Calculate_Time("read in data",start,finish);

  start=Start_Clock("berthod");

  if (reseeding) { /* create a new statefile from scratch */
    srandom(seed);
  } 

  initparam(fd_robot);

  find_path(V,voro,&finalpath);
    
  if (n>0)
    fprintf(stderr,"Optimizing the robot path\n");

  for (i=0;i<n;i++) {  
    smooth4(V,MIN_PATH_LENGTH,&finalpath);
  }

  if (curvline) {
    fprintf(stderr,"Polishing the path ...\n");
    CurvePath((configuration *)&(finalpath.d[0][0]),finalpath.dim,V);
    smooth4(V,256,&finalpath);

#ifdef DEBUG      
    for(i=0;i<finalpath.dim;i++) {
      CFK(finalpath.d[i],qp);
      if (HIT(qp,V))
	fprintf(stderr,"Collision at step%d\n",i);
    }    
#endif      
  }
  
  path_des.octet = sizeof(configuration);
  path_des.dimx = therobot.dimc[0];
  path_des.dimy = 1;
  path_des.dimz = finalpath.dim;
  
  finish=Finish_Clock("berthod");
  Calculate_Time("berthod",start,finish);
  
  fprintf(stderr,"Number of forward kinematics calls:%d\n",nb_fk_calls);
  fprintf(stderr,"Number of goodson checking calls:%d\n",nb_goodson_calls);
  fprintf(stderr,"Number of collision checking calls:%d\n",nb_hit_calls);
  fprintf(stderr,"Number of collision free calls:%d\n",nb_no_hit);

  fwrite(&path_des,sizeof(descript),1,fd_path);  
  fwrite(finalpath.d,path_des.octet,finalpath.dim,fd_path);
  fclose(fd_path);
  return(TRUE);
}

/**************************************************************************
  function to initialize the robot
***************************************************************************/

void initparam(FILE *fd_robot) {

  clock_t start,finish;
  start=Start_Clock("init parameter");
   
  xsubi[0] = 519348;
  xsubi[1] = 37465;
  xsubi[2] = 198;
  
  Read_Robot_File(fd_robot);

  Init_FK();

  finish=Finish_Clock("init parameter");
  Calculate_Time("init parameter",start,finish);
}

/**************************************************************************
  function to print help
***************************************************************************/

void PrintHelp(void) {

  fprintf(stderr,"Usage:\tsolid [-i infile] [-p pathfile] [-r robotile] [-n number of smooth][-s [seed]][-v][-c]\n");
  fprintf(stderr,"\t -i: workspace bitmap data file name [bmap.dat]\n");
  fprintf(stderr,"\t -p: output path file name [path.dat]\n");
  fprintf(stderr,"\t -r: robot description file name [robot.dat]\n");
  fprintf(stderr,"\t -n: number of smoothing steps [0]\n");
  fprintf(stderr,"\t -s: reseeding and seed\n");
  fprintf(stderr,"\t -c: polish path with curvilinear function\n");
  fprintf(stderr,"\t -v: verbose mode\n");

}

