/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <memory.h>
#include <malloc.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "synth3dp.h"

FILE *log_fd;

/*****************************************************************************
  Function to create workspace bitmap
******************************************************************************/

void Create_Bitmap(descript ws,unsigned char ***ws_bmap,int *nbseg,
		   int nbobj,segment *seg,object *world) {

  int x,y,z,n=0,i,j=0,m;
  double k,l;
  unsigned char ***obj_bmap, ***prim_bmap;
  fpoint3D fp1,fp2;
  int dimx=ws.dimx;
  int dimy=ws.dimy;
  int dimz=ws.dimz;
  double coefx=1.0/(double)dimx,coefy=1.0/(double)dimy,coefz=1.0/(double)dimz;
  boolean inside;
  fpoint3D half_dim;

  /* Contruct object and primitive bounding boxs */

#ifdef VERBOSE
  if (verbose) 
    fprintf(stderr,"Computing bounding box and line segments\n");
#endif

  for (i=0;i<nbobj;++i) {
    world[i].box.max.x = world[i].box.max.y = world[i].box.max.z = 0.0;
    world[i].box.min.x = world[i].box.min.y = world[i].box.min.z = 128.0;
    
    for (j=0;j<world[i].nbprim;j++) {

      /* initialize the bounding boxes */

      world[i].prim[j].box.max.x = 
	world[i].prim[j].box.max.y = 
	  world[i].prim[j].box.max.z=0.0;
      world[i].prim[j].box.min.x = 
	world[i].prim[j].box.min.y = 
	  world[i].prim[j].box.min.z = 128.0;

      /* fill the bounding box element along x axis */
      for(k = -0.5; k<= 0.5;k+= 1.0) {
	for(l = -0.5;l<= 0.5;l+= 1.0) {

	  seg[n].a.x = - 0.5 * world[i].prim[j].dim.x;
	  seg[n].b.x = 0.5 * world[i].prim[j].dim.x;
	  seg[n].a.y = k * world[i].prim[j].dim.y;
	  seg[n].b.y = k * world[i].prim[j].dim.y;  
	  seg[n].a.z = l * world[i].prim[j].dim.z;
	  seg[n].b.z = l * world[i].prim[j].dim.z;
	  FK_Point(world[i].prim[j].orig,seg[n].a,&(seg[n].a));
	  FK_Point(world[i].prim[j].orig,seg[n].b,&(seg[n].b));
	  
	  Update_Bbox(&(world[i].prim[j].box),seg,n);
	  Update_Bbox(&(world[i].box),seg,n);
	  n++;
	}
      }
      
      /* fill the bounding box element along y axis */
      for(k = -0.5; k<= 0.5;k+= 1.0){
	for(l = -0.5;l<= 0.5;l+= 1.0){
	  seg[n].a.x = k * world[i].prim[j].dim.x;
	  seg[n].b.x = k * world[i].prim[j].dim.x;
	  seg[n].a.y = -0.5 * world[i].prim[j].dim.y;
	  seg[n].b.y = 0.5 * world[i].prim[j].dim.y;
	  seg[n].a.z = l * world[i].prim[j].dim.z;
	  seg[n].b.z = l * world[i].prim[j].dim.z;
	  FK_Point(world[i].prim[j].orig,seg[n].a,&(seg[n].a));
	  FK_Point(world[i].prim[j].orig,seg[n].b,&(seg[n].b));
	  
	  Update_Bbox(&(world[i].prim[j].box),seg,n);
	  Update_Bbox(&(world[i].box),seg,n);
	  n++;
	}
      }
      
      /* fill the bounding box element along z axis */
      for(k = -0.5; k<= 0.5;k+= 1.0){
	for(l = -0.5;l<= 0.5;l+= 1.0){
	  
	  seg[n].a.x = k * world[i].prim[j].dim.x;
	  seg[n].b.x = k * world[i].prim[j].dim.x;
	  seg[n].a.y = l * world[i].prim[j].dim.y;
	  seg[n].b.y = l * world[i].prim[j].dim.y;
	  seg[n].a.z = - 0.5 * world[i].prim[j].dim.z;
	  seg[n].b.z = 0.5 * world[i].prim[j].dim.z;
	  FK_Point(world[i].prim[j].orig,seg[n].a,&(seg[n].a));
	  FK_Point(world[i].prim[j].orig,seg[n].b,&(seg[n].b));
	  
	  Update_Bbox(&(world[i].prim[j].box),seg,n);
	  Update_Bbox(&(world[i].box),seg,n);
	  n++;
	}
      }
    }
  }

  (*nbseg) = n;
#ifdef DEBUG
  fprintf(log_fd,"nb of segments = %d\n",n);
  if (verbose)
    fprintf(stderr,"nb of segments = %d\n",n);
#endif

  /* Convert the double bounding box into int bounding box */
  Convert_Bbox(ws,nbobj,world);

  obj_bmap = (unsigned char ***)calloc3D(dimx,dimy,dimz,sizeof(unsigned char));
  prim_bmap= (unsigned char ***)calloc3D(dimx,dimy,dimz,sizeof(unsigned char));
  
  memset(ws_bmap[0][0],FREESPACE,dimx*dimy*dimz);

#ifdef VERBOSE
  if(verbose)
    fprintf(stderr,"Filling boundary points\n");
#endif    

  /* fill the boundary as obstacle */
  m=0;

  for (x=0;x<dimx;x+=(dimx-1)) {
    for (y=0;y<dimy;y++) {
      for (z=0;z<dimz;z++) {
        ws_bmap[x][y][z]=OBSTACLE;
	m++;
      }
    }
  }
  for (y=0;y<dimy;y+=(dimy-1)) {
    for (x=0;x<dimx;x++) {
      for (z=0;z<dimz;z++) {
        ws_bmap[x][y][z]=OBSTACLE;
	m++;
      }
    }
  }
  for (z=0;z<dimz;z+=(dimz-1)) {
    for (y=0;y<dimy;y++) {
      for (x=0;x<dimx;x++) {
        ws_bmap[x][y][z]=OBSTACLE;
	m++;
      }
    }
  }

#ifdef DEBUG
  fprintf(log_fd,"Number of cells filled on boundary:%d\n",m);
  if (verbose)
    fprintf(stderr,"Number of cells filled on boundary:%d\n",m);
#endif
#ifdef VERBOSE
  if (verbose)
    fprintf(stderr,"Number of cells filled on boundary:%d\n",m);
#endif

  /* for each objects create the bitmap */

  for (i=1;i<nbobj;i++) {

#ifdef VERBOSE
    if (verbose) 
      fprintf(stderr,"Building bitmap for object #%d\n",i);
#endif

    memset(obj_bmap[0][0],0,dimx*dimy*dimz);

    /* for each primitive */
    for (j=0;j<world[i].nbprim;j++) {
#ifdef VERBOSE
    if (verbose) 
      fprintf(stderr,"Building bitmap for primitive #%d\n",j);
#endif
      memset(prim_bmap[0][0],0,dimx*dimy*dimz);
      m=0;
      half_dim.x=world[i].prim[j].dim.x*0.5;
      half_dim.y=world[i].prim[j].dim.y*0.5;
      half_dim.z=world[i].prim[j].dim.z*0.5;

      /* fill in the primitive bitmap according the bounding box */
      for (x=world[i].prim[j].ibox.min.x;x<=world[i].prim[j].ibox.max.x;x++) {
	fp1.x=x*coefx;
	for (y=world[i].prim[j].ibox.min.y;y<=world[i].prim[j].ibox.max.y;
	     y++) {
	  fp1.y=y*coefy;
	  for (z=world[i].prim[j].ibox.min.z;z<=world[i].prim[j].ibox.max.z;
	       z++) {
	    fp1.z=z*coefz;

	    BK_Point(world[i].prim[j].orig,fp1,&fp2);

	    switch (world[i].prim[j].objtype) {
	    case cube:
	      inside = (fp2.x < half_dim.x) && 	(fp2.x > -half_dim.x) && 
		(fp2.y < half_dim.y) && (fp2.y > -half_dim.y) && 
		  (fp2.z < half_dim.z) && (fp2.z > -half_dim.z);
	      break;

	    case cylinder:
	      inside=(sqrt((fp2.x*fp2.x)+(fp2.y*fp2.y))<half_dim.x)
		&&(fp2.z < half_dim.z) && (fp2.z > -half_dim.z);
	      break;

	    case sphere:
	      inside=(sqrt((fp2.x*fp2.x)+(fp2.y*fp2.y)+(fp2.z*fp2.z)) < 
		      half_dim.x);
	      break;

	    default:
	      fprintf(stderr, "No primitive class matches");
	      break;

	    }
	    
	    if (inside) {
	      m++;
	      prim_bmap[x][y][z] = 1;
	    }
	  }
	}
      }
#ifdef DEBUG
      fprintf(log_fd,"number of cells filled:%d\n",m);
    if (verbose)
      fprintf(stderr,"number of cells filled:%d\n",m);
#endif
    
      /* do operations on the primitive and the object bitmap */

      for (x=world[i].ibox.min.x;x<=world[i].ibox.max.x;x++) {
	for (y=world[i].ibox.min.y;y<=world[i].ibox.max.y;y++) {
	  for (z=world[i].ibox.min.z;z<=world[i].ibox.max.z;z++) {
	    switch (world[i].prim[j].optype) {
	    case uni:
	      if  (prim_bmap[x][y][z] == 1) obj_bmap[x][y][z] = 1;
	      break;
	    case diff:
	      if (prim_bmap[x][y][z] != obj_bmap[x][y][z]) 
		obj_bmap[x][y][z] = 1;
	      break;
	    case subs:
	      if (prim_bmap[x][y][z] == 1) 
		obj_bmap[x][y][z] = 0;
	      break;
	    default:
	      break;
	    }
	  }
	}
      }
    }
    m=0;
    for (x=world[i].ibox.min.x;x<=world[i].ibox.max.x;x++) {
      for (y=world[i].ibox.min.y;y<=world[i].ibox.max.y;y++) {
	for (z=world[i].ibox.min.z;z<=world[i].ibox.max.z;z++) {
	  if (obj_bmap[x][y][z] == 1) {
	    m++;
	    ws_bmap[x][y][z] = OBSTACLE;
	  }
	}
      }
    }
#ifdef DEBUG
    fprintf(log_fd,"number of cells filled:%d\n",m);
    if (verbose)
      fprintf(log_fd,"number of cells filled:%d\n",m);
#endif
  }

  free3D(obj_bmap);
  free3D(prim_bmap);

} /* END */

/****************************************************************************
  Update the bounding box for one object
*****************************************************************************/

void Update_Bbox(bbox *obbox, segment *seg, int i) {
  
  if (obbox->max.x < seg[i].a.x) obbox->max.x = seg[i].a.x;
  if (obbox->min.x > seg[i].a.x) obbox->min.x = seg[i].a.x;
  if (obbox->max.x < seg[i].b.x) obbox->max.x = seg[i].b.x;
  if (obbox->min.x > seg[i].b.x) obbox->min.x = seg[i].b.x;
  if (obbox->max.y < seg[i].a.y) obbox->max.y = seg[i].a.y;
  if (obbox->min.y > seg[i].a.y) obbox->min.y = seg[i].a.y;
  if (obbox->max.y < seg[i].b.y) obbox->max.y = seg[i].b.y;
  if (obbox->min.y > seg[i].b.y) obbox->min.y = seg[i].b.y;
  if (obbox->max.z < seg[i].a.z) obbox->max.z = seg[i].a.z;
  if (obbox->min.z > seg[i].a.z) obbox->min.z = seg[i].a.z;
  if (obbox->max.z < seg[i].b.z) obbox->max.z = seg[i].b.z;
  if (obbox->min.z > seg[i].b.z) obbox->min.z = seg[i].b.z;
}

/****************************************************************************
  Convert the bounding box into integer bounding box
*****************************************************************************/

void Convert_Bbox(descript ws,int nbobj, object *world) {
  
  register int i,j;
  int dimx=ws.dimx,dimy=ws.dimy,dimz=ws.dimz;

  for (i=0;i<nbobj;i++) {
    world[i].box.max.x=MIN(world[i].box.max.x+0.05,1.00);
    world[i].box.min.x=MAX(world[i].box.min.x-0.05,0.00);
    world[i].box.max.y=MIN(world[i].box.max.y+0.05,1.00);
    world[i].box.min.y=MAX(world[i].box.min.y-0.05,0.00);
    world[i].box.max.z=MIN(world[i].box.max.z+0.05,1.00);
    world[i].box.min.z=MAX(world[i].box.min.z-0.05,0.00);

    world[i].ibox.max.x = (int) (world[i].box.max.x * (double)dimx);
    world[i].ibox.min.x = (int) (world[i].box.min.x * (double)dimx);
    world[i].ibox.max.y = (int) (world[i].box.max.y * (double)dimy);
    world[i].ibox.min.y = (int) (world[i].box.min.y * (double)dimy);
    world[i].ibox.max.z = (int) (world[i].box.max.z * (double)dimz);
    world[i].ibox.min.z = (int) (world[i].box.min.z * (double)dimz);
#ifdef DEBUG
    fprintf(log_fd,"i=%d,max(x=%d,y=%d,z=%d),min(x=%d,y=%d,z=%d)\n",i,
	    world[i].ibox.max.x,world[i].ibox.max.y,
	    world[i].ibox.max.z,world[i].ibox.min.x,
	    world[i].ibox.min.y,world[i].ibox.min.z);
    if (verbose)
      fprintf(stderr,"i=%d,max(x=%d,y=%d,z=%d),min(x=%d,y=%d,z=%d)\n",i,
	      world[i].ibox.max.x,world[i].ibox.max.y,
	      world[i].ibox.max.z,world[i].ibox.min.x,
	      world[i].ibox.min.y,world[i].ibox.min.z);
#endif
    for (j=0;j<world[i].nbprim;j++) {
      world[i].prim[j].box.max.x=MIN(world[i].prim[j].box.max.x+0.05,1.00);
      world[i].prim[j].box.min.x=MAX(world[i].prim[j].box.min.x-0.05,0.00);
      world[i].prim[j].box.max.y=MIN(world[i].prim[j].box.max.y+0.05,1.00);
      world[i].prim[j].box.min.y=MAX(world[i].prim[j].box.min.y-0.05,0.00);
      world[i].prim[j].box.max.z=MIN(world[i].prim[j].box.max.z+0.05,1.00);
      world[i].prim[j].box.min.z=MAX(world[i].prim[j].box.min.z-0.05,0.00);
      
      world[i].prim[j].ibox.max.x = (int) (world[i].prim[j].box.max.x * 
					   (double)dimx);
      world[i].prim[j].ibox.min.x = (int) (world[i].prim[j].box.min.x * 
					   (double)dimx);
      world[i].prim[j].ibox.max.y = (int) (world[i].prim[j].box.max.y * 
					   (double)dimy);
      world[i].prim[j].ibox.min.y = (int) (world[i].prim[j].box.min.y * 
					   (double)dimy);
      world[i].prim[j].ibox.max.z = (int) (world[i].prim[j].box.max.z * 
					   (double)dimz);
      world[i].prim[j].ibox.min.z = (int) (world[i].prim[j].box.min.z * 
					   (double)dimz);
#ifdef DEBUG
      fprintf(log_fd,"j=%d,max(x=%d,y=%d,z=%d),min(x=%d,y=%d,z=%d)\n",j,
	      world[i].prim[j].ibox.max.x,world[i].prim[j].ibox.max.y,
	      world[i].prim[j].ibox.max.z,world[i].prim[j].ibox.min.x,
	      world[i].prim[j].ibox.min.y,world[i].prim[j].ibox.min.z);
      if (verbose)
	fprintf(log_fd,"j=%d,max(x=%d,y=%d,z=%d),min(x=%d,y=%d,z=%d)\n",j,
		world[i].prim[j].ibox.max.x,world[i].prim[j].ibox.max.y,
		world[i].prim[j].ibox.max.z,world[i].prim[j].ibox.min.x,
		world[i].prim[j].ibox.min.y,world[i].prim[j].ibox.min.z);
#endif
    }
  }
}

