/*****************************************************************************

  3D Randomized Motion Planner

  COPYRIGHT (c) 1990 
  THE BOARD OF TRUSTEES OF THE LELAND STANFORD JUNIOR UNIVERSITY
  ALL RIGHTS RESERVED

  No duplications, whole or partial, manual or electronic, may be made 
  without the express permission of the author. Any such copies, or 
  revisions thereof, must display this notice unaltered.

  Robotics Laboratory,
  Computer Science Department,
  Stanford University

  Original author: Jerome Barraquand
  Modified by Tsai-Yen Li (li@flamingo.Stanford.EDU)
  Advisor: Jean-Claude Latombe

******************************************************************************/

#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "defs.h"
#include "solidstruct.h"
#include "macro.h"
#include "xsolidp.h"
#include "parse.h"

/***************************************************************************
  declare global variables
****************************************************************************/

descript ws_des,path_des;
robot therobot;
FILE *log_fd;
int verbose;
long nb_fk_calls=0;

/***************************************************************************
  Main function for xsolid
****************************************************************************/

int main(int argc,char **argv) {
  
  char bmap_name[80]; /* File of the 3d segments */
  char seg_name[80];
  char path_name[80]; /* Path obtained */
  char robot_name[80]; /* File for robot description */
  FILE *fd_bmap,*fd_seg,*fd_path, *fd_robot;
  segment *seg;
  int nbseg;
  float f,ty,tz,theta,phi,d;
  path X;
  struct arg *args;
  
  args=parse_args(argc,argv);

  if (args[chindex('h')].ch) {
    PrintHelp();
    exit(0);
  }
  if (args[chindex('v')].ch) 
    verbose=1;
  else
    verbose=0;

  GetFloatOpt(args,'f',&f,0.05,0.0,1.0);
  GetFloatOpt(args,'d',&d,1.5,0,10.0);
  GetFloatOpt(args,'t',&theta,(1.0/6.0),0.0,1.0);
  GetFloatOpt(args,'w',&phi,(1.0/12.0),0.0,1.0);
  GetFloatOpt(args,'y',&ty,0.07,0.0,1.0);
  GetFloatOpt(args,'z',&tz,0.07,0.0,1.0);
  fprintf(stderr,"f=%f,d=%f,theta=%f,phi=%f,ty=%f,tz=%f\n",
	  f,d,theta,phi,ty,tz);

  theta *= 2*M_PI;
  phi *= 2*M_PI;

  if (!GetStrOpt(args,'i',bmap_name,"bmap.dat",0,12)) 
    fprintf(stderr,"Use default workspace bitmap file name %s\n",bmap_name);
  if (!GetStrOpt(args,'r',robot_name,"robot.dat",0,12)) 
    fprintf(stderr,"Use default robot data file name %s\n",robot_name);
  if (!GetStrOpt(args,'p',path_name,"path.dat",0,12)) 
    fprintf(stderr,"Use default path file name %s\n",path_name);

  strcpy(seg_name,bmap_name);
  strcat(seg_name,".seg");

  if(!(fd_bmap = fopen(bmap_name,"r"))) {
    fprintf(stderr,"Cannot open file %s\n",bmap_name);	
    exit(-1);
  }

  if(!(fd_seg = fopen(seg_name,"r"))) {
    fprintf(stderr,"Cannot open file %s\n",seg_name);	
    exit(-1);
  }

  if(!(fd_robot = fopen(robot_name,"r"))) {
    fprintf(stderr,"Cannot open file %s\n",robot_name);	
    exit(-1);
  }
  
  if(!(fd_path = fopen(path_name,"r"))) {
    fprintf(stderr,"Cannot open path %s\n",path_name);
    exit(-1);
  }

  fread(&ws_des,sizeof(descript),1,fd_bmap);
  fread(&nbseg,sizeof(int),1,fd_seg);
  seg = (segment *)calloc(nbseg,sizeof(segment));
  fread(seg,sizeof(segment),nbseg,fd_seg);
  fread(&path_des,sizeof(descript),1,fd_path);

  X.dim = path_des.dimz;
  X.d = (configuration *)calloc(X.dim,sizeof(configuration));
  fread(X.d,sizeof(configuration),X.dim,fd_path);

  ws_des.dimx = ws_des.dimy = ws_des.dimz = 128;

  initparam(fd_robot);

  visuel(X,seg,nbseg,f,d,theta,phi,ty,tz);

  fclose(fd_seg);
  fclose(fd_bmap);
  fclose(fd_robot);
  fclose(fd_path); 
  return(TRUE);
}


/***************************************************************************
  Function to initialize the robot parameter
****************************************************************************/

void initparam(FILE *fd_robot) {

  Read_Robot_File(fd_robot);
}

/***************************************************************************
  Function to print help
****************************************************************************/

void PrintHelp(void) {

  fprintf(stderr,
	  "Usage: xsolid [-f f][-d d][-t theta][-w phi][-y ty][-z tz][-i infile] [-p pathfile][-r robotfile][-v]\n");
  fprintf(stderr,"\t -i: workspace bitmap data file name [bmap.dat]\n");
  fprintf(stderr,"\t -p: output path file name [path.dat]\n");
  fprintf(stderr,"\t -r: robot description file name [robot.dat]\n");
  fprintf(stderr,"\t -v: verbose mode\n");
}
